------------------------------------------------------------------------
-- Adds spine indices to types
------------------------------------------------------------------------

-- # import Explicit.Types

infixl 70 /Sp
infixl 45 ∷Sp

------------------------------------------------------------------------
-- Spines

-- This is (basically) what you get when you drop the contents of the
-- El constructors.

data Spine : Set where
  ✶Sp  : Spine
  ΠSp  : Spine -> Spine -> Spine
  ElSp : Spine

------------------------------------------------------------------------
-- / preserves spines

-- Here /Sp is defined, a version of / whose type ensures spine
-- preservation.
--
-- I could have indexed Ty on a Spine. That way /Sp would have been
-- unnecessary, because the type of / would have ensured spine
-- preservation. However, adding an index to a widely used type means
-- modifying a huge number of type signatures, so it is a very
-- invasive procedure.
--
-- In this case I just need the spine preservation property in a few
-- places, so I follow the following, cheaper, path:
--
-- 1) Define a relation (TySp) between types and spines.
--
-- 2) Define a function (toTySp) for converting types to elements of
--    the new relation.
--
-- 3) Port relevant functions (/) to this relation.

mutual

  TySp (Γ : Ctxt) : Spine -> Ty Γ -> Set
  TySp = TySp' _

  data TySp' : (Γ : Ctxt) -> Spine -> Ty Γ -> Set where
    ✶TySp : (Γ : Ctxt) -> TySp ✶Sp (✶ Γ)
    ΠTySp (s₁, s₂ : Spine) (Γ : Ctxt) (τ₁ : Ty Γ) (τ₂ : Ty (Γ ▻ τ₁))
      : TySp s₁ τ₁ -> TySp s₂ τ₂ -> TySp (ΠSp s₁ s₂) (Π τ₁ τ₂)
    ElTySp (Γ : Ctxt) : (t : Γ ⊢ ✶ Γ) -> TySp ElSp (El t)

spine (Γ : Ctxt) : Ty Γ -> Spine
spine (✶ Γ)     = ✶Sp
spine (Π τ₁ τ₂) = ΠSp (spine τ₁) (spine τ₂)
spine (El t)    = ElSp

tyToTySp (Γ : Ctxt) : (τ : Ty Γ) -> TySp (spine τ) τ
tyToTySp (✶ Γ)     = ✶TySp Γ
tyToTySp (Π τ₁ τ₂) = ΠTySp (tyToTySp τ₁) (tyToTySp τ₂)
tyToTySp (El t)    = ElTySp t

(/Sp) (s : Spine) (Γ, Δ : Ctxt) (τ : Ty Γ)
  : TySp s τ -> (ρ : Subst Γ Δ) -> TySp s (τ / ρ)
✶TySp Γ       /Sp ρ = ✶TySp _
ΠTySp τs₁ τs₂ /Sp ρ = ΠTySp (τs₁ /Sp ρ) (τs₂ /Sp ρ ↑ _)
ElTySp t      /Sp ρ = ElTySp (t /⊢ ρ)

------------------------------------------------------------------------
-- Equal types have equal spines

-- This function is related to the tyEqSpineEq lemma that I would have
-- added if I had indexed types on spines.

(∷Sp) (s : Spine)
      (Γ¹ : Ctxt) (τ¹ : Ty Γ¹)
      (Γ² : Ctxt) (τ² : Ty Γ²)
  : TySp s τ¹ -> τ¹ ≛ τ² -> TySp s τ²
✶TySp Γ¹      ∷Sp ✶Cong eqΓ       = ✶TySp _
ΠTySp τs₁ τs₂ ∷Sp ΠCong eqτ₁ eqτ₂ = ΠTySp (τs₁ ∷Sp eqτ₁) (τs₂ ∷Sp eqτ₂)
ElTySp t¹     ∷Sp ElCong eqt      = ElTySp _
